/**************************************************************************

NAME
   client.c -- support library for DCP client programs

DESCRIPTION
   See the dcpc(3) manual page.  Note: This implementation is restricted
to having only one DCP connection open at a time.

AUTHOR
   Eric S. Raymond <esr@thyrsus.com>

**************************************************************************/
#include "config.h"
#include <ctype.h>
#include "socket.h"
#include "dcplib.h"
#include "dcpc.h"
#include "dcperror.h"

int dcp_errno;

static int sock = -1;		/* the socket connection */
static char buf[DCPBUFSIZE];

static int state;

int dcp_open(const char *server, int portnum)
/* open a DCP connection to a server */
{
    return((sock = SockOpen(server, portnum)));
}

int dcp_close(void)
/* close your server connection */
{
    int	ok;

    gen_send(sock, "QUIT");
    if (gen_recv(sock, buf, sizeof(buf)))
	return(DCP_ERR);
    close(sock);
    sock = -1;
    state = STATE_INITIAL;
}

int dcp_hello(const char *userid)
/* identify yourself to the server */
{
    int ok;

    gen_send(sock, "HELO %s", userid);
    if (gen_recv(sock, buf, sizeof(buf)))
	return(DCP_ERR);
    state = STATE_IDENTIFIED;
    /* fall through */
}

int dcp_select(const char *project, const char *clientid, char *secret)
/* get authenticated for a particular project */
{
    switch(state)
    {
    case STATE_INITIAL:
	dcp_errno = EPROTERR;
	return(DCP_ERR);

    case STATE_AUTHENTICATED:
	/*
	 * Authenticated to another project?  Then we're switching projects -- 
	 * just fall through as though we're in STATE_IDENTIFIED.
	 */

    case STATE_IDENTIFIED:
	gen_send(sock, "PROJECT %s %s", project, clientid);
	if (gen_recv(sock, buf, sizeof(buf)))
	    return(DCP_ERR);
	state = STATE_IDENTIFIED;
	/* fall through */

    case STATE_SELECTED:
	gen_send(sock, "AUTH ONETIME %s", secret);
	if (gen_recv(sock, buf, sizeof(buf)))
	    return(DCP_ERR);
	else
	{
	    char	*cp;

	    state = STATE_SELECTED;
	    for (cp = buf + 2; *cp && *cp != '\r' && *cp != '\n'; cp++)
		*secret++ = *cp;
	    *secret = '\0';
	    return(gen_recv(sock, buf, sizeof(buf)));
	}
	break;
    }

    return(DCP_OK);
}

static void entry_parse(const char *buf, struct dcp_entry *entry)
/* parse a line of DCP event specifications */
{
    struct dcp_event *ep;
    register char *tp;
    register const char *cp = buf;
    int actual_events = 0;

    memset(&entry, '\0', sizeof(entry));

    while (!isdigit(*cp))
	cp++;

    entry->nevents = atoi(cp);

    while (isdigit(*cp))
	cp++;

    for (ep = entry->events; ep <= entry->events + entry->nevents; ep++)
    {
	while (*cp && isspace(*cp))
	    cp++;

	if (*cp != '<')
	{
	    dcp_errno = EBADFRMT;
	    return;
	}

	for (tp = ep->userid; *cp && *cp != ':'; cp++)
	    *tp++ = *cp;

	if (*cp == ':')
	    *tp = '\0';
	else
	{
	    dcp_errno = EBADFRMT;
	    return;
	}

	for (tp = ep->userid; *cp && *cp != '>'; cp++)
	    *tp++ = *cp;

	if (*cp != '>')
	    *tp = '\0';
	else
	{
	    dcp_errno = EBADFRMT;
	    return;
	}

	if (*cp == '+')
	{
	    ep->type = EVENT_RESERVE;
	    ep->time = atol(cp);

	    while (isdigit(*cp))
		cp++;
	    --cp;
	}

	if (*cp == '=')
	{
	    ep->type = EVENT_REPORT;
	    ++cp;
	    for (tp = ep->userid; *cp; cp++)
		if (*cp == '"')
		    break;
		else
		    *tp++ = *cp;
	    *tp = '\0';
	    ++cp;
	}
    }
}

static struct dcp_entry *dcp_single_op(const char *fmt, const char *key)
/* send down a command, expect a single line of event-data response */
{
    static struct dcp_entry res;

    if (state != STATE_SELECTED)
    {
	dcp_errno = EPROTERR;
	return((struct dcp_entry *)NULL);
    }

    gen_send(sock, fmt, key);
    if (gen_recv(sock, buf, sizeof(buf)) || atoi(buf+2) != 1)
	return((struct dcp_entry *)NULL);

    /* get the actual response line */
    if (gen_recv(sock, buf, sizeof(buf)))
	return((struct dcp_entry *)NULL);

    /* now interpret it */
    entry_parse(buf, &res);

    /* eat the trailing response line */
    if (gen_recv(sock, buf, sizeof(buf)))
    {
	dcp_errno = EPROTERR;
	return((struct dcp_entry *)NULL);
    }

    return(&res);
}

static int dcp_range_op(const char *fmt,
    int min, int max, int *count,
    void (*hook)(const char *, struct dcp_entry *))
{
    int nkeys;

    if (state != STATE_SELECTED)
    {
	dcp_errno = EPROTERR;
	return(DCP_ERR);
    }

    gen_send(sock, fmt, min, max);
    if (gen_recv(sock, buf, sizeof(buf)) || (*count = atoi(buf+2)) == 0)
	return(DCP_ERR);

    for (nkeys = *count; nkeys; nkeys--)
    {
	char	*key, *report;
	struct dcp_entry res;

	/* get the actual response line */
	if (gen_recv(sock, buf, sizeof(buf)))
	    return(DCP_ERR);

	if (buf[0] != '*')
	{
	    dcp_errno = EPROTERR;
	    return(DCP_ERR);
	}

	key = nexttok(buf + 2);
	report = nexttok(NULL);

	memset(&res, '\0', sizeof(res));

	if (buf[2] == 'O' && buf[3] == 'K')
	    entry_parse(buf + 4, &res);

	if (hook)
	    (*hook)(key, &res);
    }

    /* eat the trailing response line */
    if (gen_recv(sock, buf, sizeof(buf)))
    {
	dcp_errno = EPROTERR;
	return(DCP_ERR);
    }

    return(DCP_OK);
}

struct dcp_entry *dcp_view_key(const char *key)
/* retrieve data associated with a string key */
{
    return(dcp_single_op("VIEW %s", key));
}

int dcp_view_range(int min, int max, int *count,
                   void (*hook)(const char *, struct dcp_entry *))
/* feed data associated with given range to a hook function */
{
    return(dcp_range_op("VIEW %d-%d", min, max, count, hook));
}

struct dcp_entry *dcp_request_key(const char *key)
/* request checkout of a string key */
{
    return(dcp_single_op("REQUEST %s", key));
}

int dcp_request_range(int min, int max, int *count,
                   void (*hook)(const char *, struct dcp_entry *))
/* request checkout of a numeric key range */
{
    return(dcp_range_op("VIEW %d-%d", min, max, count, hook));
}

struct dcp_entry *dcp_release_key(const char *key)
/* release a string key without entering a report */
{
    return(dcp_single_op("RELEASE %s", key));
}

int dcp_release_range(int min, int max, int *count,
                   void (*hook)(const char *, struct dcp_entry *))
/* release a range of numeric keys without entering a report */
{
    return(dcp_range_op("RELEASE %d-%d", min, max, count, hook));
}

int dcp_report_key(const char *key, const char *data)
/* report data on a string key */
{
    if (state != STATE_SELECTED)
    {
	dcp_errno = EPROTERR;
	return(DCP_ERR);
    }

    gen_send(sock, "REPORT %s %s", key, data);
    if (gen_recv(sock, buf, sizeof(buf)))
	return(DCP_ERR);
    else
	return(DCP_OK);
}

/* dcpc.c ends here */
