/***************************************************************************

NAME
    entry.c -- database-access layer for DCP server and administrative tools

DESCRIPTION
    These functions hide the messy details of GDBM database access, giving
us records in the format we want.

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <gdbm.h>
#include <time.h>
#include <stdarg.h>
#include "dcpd.h"

/****************************************************************************
 *
 * Database-access layer
 *
 ****************************************************************************/

/*
 * This is the on-disk representation of a struct cp_entry
 * It's designed for space economy -- there might be millions of these!
 */
struct dcp_event_flattened
{
    time_t	time;		/* reservation timestamp */
    short	type;		/* event type */
    short	userid_len;	/* length of user IDs in following data */
    short	clientid_len;	/* length of user IDs in following data */
    short	report_len;	/* length of properties in following data */
};

static char *gdbm_fatalerr;

static void fatalerr(char *str)
/* hook function for notification of GDBM fatal errors */
{
    gdbm_fatalerr = str;
}

GDBM_FILE entry_open(const char *base)
/* open a GDBM database */
{
    char	buf[BUFSIZ];

    sprintf(buf, "%s/%s-db", BASEDIR, base);
    return(gdbm_open(buf, 0, GDBM_WRCREAT, 0666, fatalerr));
}

struct dcp_entry *entry_get(GDBM_FILE db, const char *key)
/* return allocated, unpacked copy of data associated with given key */
{
    static struct dcp_entry frame;
    datum keydat, res;
    int i, offset;

    keydat.dsize = strlen(key);
    keydat.dptr = (char *)key;
    res = gdbm_fetch(db, keydat);
    if (!res.dptr)
	memset(&frame, '\0', sizeof(frame));
    else
    {
	char *data = res.dptr;

	frame.nevents = *((short *) data);
	data += sizeof(short);

	for (offset = i = 0; i < frame.nevents; i++)
	{
	    struct dcp_event_flattened *flat = (struct dcp_event_flattened *)data;

	    frame.events[i].type = flat->type;
	    frame.events[i].time = flat->time;
	    offset += sizeof(struct dcp_event_flattened);

	    strcpy(frame.events[i].userid, data + offset);
	    offset += flat->userid_len;
	    strcpy(frame.events[i].clientid, data + offset);
	    offset += flat->clientid_len;
	    strcpy(frame.events[i].report, data + offset);
	    offset += flat->report_len;
	}
	free(res.dptr);
    }

    return(&frame);
}

int entry_put(GDBM_FILE db, const char *key, struct dcp_entry *en)
/* write given entry to the database */
{
    static struct dcp_event_flattened *flat;
    char *data;
    datum keydat, res;
    int size, offset, i;

    size = sizeof(short);
    for (i = 0; i < en->nevents; i++)
    {
	size += sizeof(struct dcp_event_flattened);
	size += (strlen(en->events[i].userid) + 1);
	size += (strlen(en->events[i].clientid) + 1);
	size += (strlen(en->events[i].report) + 1);
    }

    if ((data = malloc(size)) == (char *)NULL)
	return(-1);

    (*(short *)data) = en->nevents;
    offset = sizeof(short);
    for (i = 0; i < en->nevents; i++)
    {
	flat = (struct dcp_event_flattened *)(data + offset);
	flat->time = en->events[i].time;
	flat->type = en->events[i].type;
	offset += sizeof(struct dcp_event_flattened);

	strcpy(data + offset, en->events[i].userid);
	flat->userid_len = strlen(en->events[i].userid) + 1;
	offset += flat->userid_len;

	strcpy(data + offset, en->events[i].clientid);
	flat->clientid_len = strlen(en->events[i].clientid) + 1;
	offset += flat->clientid_len;

	strcpy(data + offset, en->events[i].report);
	flat->report_len = strlen(en->events[i].report) + 1;
	offset += flat->report_len;
    }

    keydat.dsize = strlen(key);
    keydat.dptr = (char *)key;
    res.dsize = size;
    res.dptr = data;
    gdbm_store(db, keydat, res, GDBM_REPLACE);
    free(data);

    return(0);
}

void entry_print(struct dcp_service *service, struct dcp_entry *en, FILE *out)
/* print an invertible representation of events associated with an entry */
{
    int i;

    /* print counts first, it may be all the reader needs */
    fprintf(out, "%d ", en->nevents);

    for (i = 0; i < en->nevents; i++)
	switch (en->events[i].type)
	{
	case EVENT_RESERVE:
	    fprintf(out, "<%s:%s>+%d", 
		    en->events[i].userid,
		    en->events[i].clientid,
		    en->events[i].time);
	    break;

	case EVENT_REPORT:
	    fprintf(out, "<%s:%s>=\"%s\" ", 
		    en->events[i].userid, en->events[i].clientid,
		    (*service->unpack)(en->events[i].report));
	    break;

	default:
	    break;
	}
}

/****************************************************************************
 *
 * Authentication handling
 *
 ****************************************************************************/

int dcp_getauth(char *base, char *userid, char *secret)
/* check an authorization key and return the next one */
{
    char	buf[BUFSIZ];
    datum 	keydat, res;
    GDBM_FILE	db;
    int		authorization;

    sprintf(buf, "%s/%s-auth", BASEDIR, base);
    if ((db = gdbm_open(buf, 0, GDBM_WRCREAT, 0666, fatalerr)) == NULL)
	return(DCP_ERR);

    keydat.dsize = strlen(userid) + 1;
    keydat.dptr = (char *)userid;

    res = gdbm_fetch(db, keydat);

    if (res.dptr == (char *)NULL)
	authorization = DCP_OK;
    else if (strcmp(secret, res.dptr)) 
    {
	free(res.dptr);
	authorization = DCP_ERR;
    }
    else
    {
	free(res.dptr);
	authorization = DCP_OK;
    }

    if (authorization == DCP_OK)
    {
	res.dptr = passgen();
	res.dsize = strlen(res.dptr) + 1;
	if (gdbm_store(db, keydat, res, GDBM_REPLACE) == 0)
	    strcpy(secret, res.dptr);
    }

    gdbm_close(db);

    return(authorization);
}

/****************************************************************************
 *
 * Log file handling
 *
 ****************************************************************************/

static FILE *logfp;

void log_init(const char *base)
/* initialize logging for given project */
{
    char	buf[BUFSIZ];

    if (logfp)
	fclose(logfp);

    sprintf(buf, "%s/%s-log", BASEDIR, base);
    logfp = fopen(buf, "a");
}

void log_transact(char *fmt, ...)
/* log a transaction entry */
{
    va_list ap;

    va_start(ap, fmt);
    vfprintf(logfp, fmt, ap);
    va_end(ap);
}

void log_close(void)
/* close the log */
{
    fclose(logfp);
}

/* entry.c ends here */


