/***************************************************************************

NAME
    server.c -- server state machine for DCP protocol

DESCRIPTION
    The DCP protocol server state machine itself.

AUTHOR
   Eric S. Raymond <esr@thyrsus.com>

***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <assert.h>
#include <gdbm.h>
#include "dcpd.h"

/* globals available once we're in IDENTIFIED state */
static char	userid[BUFSIZ];		/* user/host ID */
static char	clientid[BUFSIZ];	/* client program ID */

/****************************************************************************
 *
 * Command parsing
 *
 ****************************************************************************/

static process_range(GDBM_FILE db,
		     const char *key,
		     struct dcp_service *service,
		     int (*hook)(const char *,struct dcp_service *,struct dcp_entry*),
		     FILE *out)
/* process a key or given range of keys with a specified hook */
{
    char	buf2[BUFSIZ];
    int	i, lower, upper, vcount = 0, errcount = 0;

    if (sscanf(key, "%d-%d", &lower, &upper) == 2)
    {
	struct dcp_entry *en;

	for (i = lower; i <= upper; i++)
	{
	    sprintf(buf2, "%d", i);
	    if ((service->key_ok)(buf2))
		vcount++;
	}
	fprintf(out, "* %d keys\r\n", vcount);

	for (i = lower; i <= upper; i++)
	{
	    sprintf(buf2, "%d", i);
	    if (!(service->key_ok)(buf2))
		continue;

	    en = entry_get(db, buf2);

	    /* caller may want to alter the key */
	    if (!hook)
		fprintf(out, "* OK \"%s\" ", buf2);
	    else
	    {
		if (hook(buf2, service, en) || entry_put(db, buf2, en))
		{
		    errcount++;
		    fprintf(out, "* ERR \"%s\" ", buf2);
		}
		else
		    fprintf(out, "* OK \"%s\"" , buf2);
	    }

	    entry_print(service, en, out);
	    fprintf(out, " %d\r\n", (service->estimate)(key));
	}

	if (errcount)
	{
	    fprintf(out, "ERR 015 EGDBMWR Database write error.\r\n");
	    return;
	}
    }
    else if (!(service->key_ok)(key))
    {
	fprintf(out, "ERR 012 EKEYINV Invalid key.\r\n");
	return;
    }
    else
    {
	struct dcp_entry *en = entry_get(db, key);
	int error = 0;

	vcount = 1;

	/* caller may want to alter the key */
	if (!hook)
	    fprintf(out, "* OK \"%s\" " , key);
	else
	{
	    if (hook(key, service, en) || entry_put(db, key, en))
	    {
		error = 1;
		fprintf(out, "* ERR \"%s\" ", key);
	    }
	    else
		fprintf(out, "* OK \"%s\" ", key);
	}

	entry_print(service, en, out);
	fprintf(out, " %d\r\n", (service->estimate)(key));

	if (error)
	{
	    fprintf(out, "ERR 015 EGDBMWR Database write error.\r\n");
	    return;
	}
    }
    
    fprintf(out, "OK %s %d key(s)\r\n", service->name, vcount);
}

static int lock_hook(const char *key,
		      struct dcp_service *service,
		      struct dcp_entry *en)
/* add a reservation to the given record */
{
    int i, status;
    time_t now;

    time(&now);

    for (i = 0; i < en->nevents; i++)
	if (!strcmp(userid, en->events[i].userid))
	{
	    log_transact("%ld, \"%s\": <%s:%s> request failed (duplicate)\n",
			 now, key, userid, clientid);
	    return(-1);
	}

    /* don't step on preexisting expired events unless necessary */
    if (en->nevents < MAX_EVENTS)
    {
	en->events[en->nevents].time = now;
	en->events[en->nevents].type = EVENT_RESERVE;
	strcpy(en->events[en->nevents].userid, userid);
	strcpy(en->events[en->nevents].clientid, clientid);
	en->events[en->nevents].report[0] = '\0';
	en->nevents++;
	log_transact("%ld, \"%s\": <%s:%s> request succeeded\n",
		     now, key, userid, clientid);
	return(0);
    }
    else
    {
	int	oldi;
	time_t	age, oldest = 0;

	for (i = 0; i < en->nevents; i++)
	{
	    age = (now - en->events[i].time);
	    if (age > service->expiry && age > oldest)
	    {
		oldest = age;
		oldi = i;
	    }
	}

	if (oldi)
	{
	    log_transact("%ld, \"%s\": <%s:%s> request replaced\n",
			 now, key, 
			 en->events[oldi].userid,
			 en->events[oldi].clientid);
	    log_transact("%ld, \"%s\": <%s:%s> request succeeded\n", 
			 now, key, 
			 userid,
			 clientid);
	    en->events[oldi].time = now;
	    en->events[oldi].type = EVENT_RESERVE;
	    strcpy(en->events[oldi].userid, userid);
	    strcpy(en->events[oldi].clientid, clientid);
	    return(0);
	}
    }

    log_transact("%ld, \"%s\": <%s:%s> request failed (too many)\n",
		 now, key, userid, clientid);
    return(-1);
}

static int unlock_hook(const char *key,
		      struct dcp_service *service,
		      struct dcp_entry *en)
/* remove a reservation from the given record */
{
    time_t now;
    int i, j;

    time(&now);
    for (i = 0; i < en->nevents; i++)
	if (!strcmp(userid, en->events[i].userid) && !strcmp(clientid, en->events[i].clientid))
	{
	    for (j = i + 1; j < en->nevents; j++)
		en->events[i++] = en->events[j];
	    en->nevents--;
	    log_transact("%ld, \"%s\": <%s:%s> release succeeded\n",
			 now, key, userid, clientid);
	    return(0);
	}

    log_transact("%ld, \"%s\": <%s> release failed\n", now, key, userid);
    return(-1);
}

#define prefix(buf, pref)	!strncasecmp(buf, pref, strlen(pref))

int dcp_server(int admin, FILE *in, struct dcp_service **services, FILE *out)
/* state machine for server side of DCP protocol */
{
    struct dcp_service *service;
    char	buf[BUFSIZ], buf2[BUFSIZ], key[BUFSIZ], *cp;
    char	method[BUFSIZ];
    int		state, st;
    GDBM_FILE	db;

    /* we'll need the hostname for the greeting line */
    if (gethostname(buf, sizeof(buf) - 1))
    {
	perror("dcpd");
	exit(1);
    }

    if (admin)
	fprintf(out, "[Administrative mode: extended commands enabled]\n");

    /* utter the greeting line */
    fprintf(out, "OK DCP %s %s <%s>\r\n", DCP_SERVER_VERSION, buf, "FOO");

    if (admin)
    {
	state = STATE_IDENTIFIED;
	strcpy(userid, getlogin());
    }
    else
    {
	state = STATE_INITIAL;
	userid[0] = '\0';		/* in case we QUIT before HELO */
    }
    db = (GDBM_FILE)NULL;

    srand(time((time_t *)NULL) + getpid());

    while (admin && fputs(">", out), fgets(buf, sizeof(buf) - 1, in))
    {
	/*
	 * Administration commands
	 */
	if (admin)
	{
	    datum key, data;

	    switch(buf[0])
	    {
	    case '?':
#define P(s) fputs(s, out)
		P("Here are the administrative commands:\n");
		P("q                -- quit (like QUIT)\n");
		P("p <name>         -- join project <name> (like PROJECT)\n");

		P("y                -- list database in key-report format\n"); 
#undef P
		if (db)
		    fprintf(out, "Project %s is selected\n", service->name);
		continue;
	    case 'y':
		if (!db)
		{
		    fputs("No project is selected\n", out);
		    continue;
		}
		key = gdbm_firstkey(db);
		do {
		    char	keybuf[MAX_KEYLEN];

		    memcpy(keybuf, key.dptr, key.dsize);
		    keybuf[key.dsize] = '\0';
		    free(key.dptr);

		    fprintf(out, "\"%s\" ", keybuf);
		    entry_print(service, entry_get(db, keybuf), out);
		    putc('\n', out);
		} while
		    ((key = gdbm_nextkey(db, key)).dptr);
		continue;
	    }
	}

	/* a motion to adjourn is always in order */
	if (!strncasecmp(buf, "QUIT", 4) || (admin && buf[0] == 'q'))
	{
	    fprintf(out, "OK Goodbye %s, please visit again.\r\n", userid);
	    break;
	}

	/* we may need a HELO before proceeding */
	if (state == STATE_INITIAL)
	{
	    if (prefix(buf, "HELO") && (cp = nexttok(buf + 4)))
	    {
		strcpy(userid, cp);

		fprintf(out, "OK Hello %s, pleased to meet you.\r\n",userid);
		state = STATE_IDENTIFIED;
	    }
	    else
		fprintf(out, "ERR 001 ENOHELO I'm expecting HELO.\r\n");
	    continue;
	}

	assert(state >= STATE_IDENTIFIED);

	/* we can accept a PROJECT now */
	if (prefix(buf, "PROJECT") && (cp = nexttok(buf + 7))
	    || (admin && buf[0] == 'p' && (cp = nexttok(buf + 1))))
	{
	    struct dcp_service **svc;

	    for (svc = services; *svc; svc++)
		if (!strcasecmp((*svc)->name, cp))
		{
		    GDBM_FILE	newdb;

		    service = *svc;

		    /* use system default block size */
		    if ((newdb = entry_open(service->file)) == NULL)
		    {
			fprintf(out,
				"ERR 002 EPROPEN Couldn't open database.\r\n");
			goto openerr;
		    }

		    fprintf(out, "OK %s %s\r\n", service->name, service->desc);
		    if (db)
			dbm_close(db);
		    db = newdb;
		    log_init(service->file);
		    goto project_ok;
		}
	    fprintf(out, "ERR 003 ENOSUCH No such database here.\r\n");
	openerr:
	    continue;
	project_ok:
	    if (admin)
		state = STATE_AUTHENTICATED;
	    else
		state = STATE_SELECTED;

	    if ((cp = nexttok(NULL)))
		strcpy(clientid, cp);
	    else
		strcpy(clientid, "*");
	    continue;
	}
	else if (state == STATE_IDENTIFIED)
	{
	    fprintf(out, "ERR ENOPROJ 004 I'm expecting PROJECT.\r\n");
	    continue;
	}

	assert(state == STATE_SELECTED || state == STATE_AUTHENTICATED);

	/* we may accept an AUTH at this point, but it's not required */
	if (state == STATE_SELECTED && prefix(buf, "AUTH"))
	{
	    int auth_ok = 0;

	    if ((cp = nexttok(buf + 4)) == (char *)NULL)
	    {
		fprintf(out, "ERR 005 EAUTHSX AUTH syntax error after AUTH.\r\n");
		continue;
	    }
	    else if (strcmp(cp, "ONETIME"))
	    {
		fprintf(out, "ERR 006 EAUTHBD Unknown authentication method.\r\n", cp);
		continue;
	    }

	    if ((cp = nexttok(NULL)) == (char *)NULL)
	    {
		fprintf(out, "ERR 007 EAUTHMN AUTH syntax error after method.\r\n");
		continue;
	    }
	    else 
		strcpy(key, cp);

	    if (dcp_getauth(service->file, userid, key) == DCP_OK)
	    {
		fprintf(out, "* %s\r\n", key);
		fprintf(out, "OK %s authenticated.\r\n", userid);
		state = STATE_AUTHENTICATED;
	    }
	    else
	    {
		time_t	now;

		fprintf(out, "ERR 008 EBADAUT Authentication failed.\r\n");
		time(&now);
		log_transact("%ld: <%s:%s> authentication failure\n",
			 now, userid, clientid);
	    }
	    continue;
	}

	/* now the commands that do the real work */

	/* record data into the database */
	if (prefix(buf, "REPORT"))
	{
	    struct dcp_entry *en;
	    int	i;
	    time_t now;

	    if (state != STATE_AUTHENTICATED)
	    {
		fprintf(out,
		    "ERR 009 ENOAUTH You can't do REPORT before authenticating.\r\n");
		continue;
	    }

	    for (cp = buf; isalpha(*cp); cp++)
		continue;
	    if ((cp = nexttok(cp)) == (char *)NULL)
	    {
		fprintf(out, "ERR 010 ERECOSX REPORT syntax error after REPORT.\r\n");
		continue;
	    }
	    else 
		strcpy(key, cp);

	    if ((cp = nexttok(NULL)) == (char *)NULL)
	    {
		fprintf(out, "ERR 011 ERECOKY REPORT syntax error after key.\r\n");
		continue;
	    }
	    else 
		strcpy(buf2, cp);

	    /* syntactically we have a key/value pair now -- check semantics */

	    /* check that the key is valid */
	    if (!service->key_ok(key))
	    {
		fprintf(out, "ERR 012 EKEYINV Invalid key.\r\n");
		continue;
	    }

	    /* can't REPORT data unless you have a reservation on it */
	    en = entry_get(db, key);
	    for (i = 0; i < en->nevents; i++)
		if (!strcmp(en->events[i].userid, userid) 
				&& !strcmp(en->events[i].clientid, clientid))
		    goto id_ok;
	    fprintf(out, "ERR 013 ENORESV You have no reservation on this key.\r\n");
	    continue;
	id_ok:;
	    
	    /* now validate the data */
	    if ((st = service->report_ok(key, en, buf2)) < 0)
	    {
		fprintf(out, "ERR 014 EDATINV Invalid data.\r\n");
		continue;
	    }
	    else
		strncpy(en->events[i].report, buf2, sizeof(en->events[i].report)-1);

	    /* write out the new entry */
	    time(&now);
	    if (entry_put(db, key, en))
	    {
		fprintf(out, "ERR 015 EGDBMWR Database write error.\r\n");
		continue;
	    }

	    /* confirm transaction to client */
	    fprintf(out, "OK %s \"%s\" <= \"%s\"\r\n",
		    service->name, key, buf2);
	    
	    /* log the transaction */
	    log_transact("%ld: <%s:%s> gave key \"%s\" report \"%s\"\n",
			 now, userid, clientid, key, buf2);
	    continue;
	}

	/* view one or more keys without locking */
	if (prefix(buf, "VIEW") || (admin && buf[0] == 'v'))
	{
	    for (cp = buf; isalpha(*cp); cp++)
		continue;
	    if ((cp = nexttok(cp)) == (char *)NULL)
	    {
		fprintf(out, "ERR 016 EVIESXQ VIEW syntax error after VIEW.\r\n");
		continue;
	    }
	    else 
		strcpy(key, cp);

	    process_range(db, key, service, NULL, out);
	    continue;
        }

	/* request one or more keys with locking */
	if (prefix(buf, "REQUEST"))
	{
	    for (cp = buf; isalpha(*cp); cp++)
		continue;
	    if ((cp = nexttok(cp)) == (char *)NULL)
		(*service->key_gen)(key);
	    else 
		strcpy(key, cp);

	    process_range(db, key, service, lock_hook, out);
	    continue;
        }

	/* release a lock on a previously requested key */
	if (prefix(buf, "RELEASE"))
	{
	    for (cp = buf; isalpha(*cp); cp++)
		continue;
	    if ((cp = nexttok(cp)) == (char *)NULL)
	    {
		fprintf(out, "ERR 017 ERELSXR RELEASE syntax error after RELEASE.\r\n");
		continue;
	    }
	    else 
		strcpy(key, cp);

	    process_range(db, key, service, unlock_hook, out);
	    continue;
        }

	/* we didn't match any valid commands */
	fprintf(out, "ERR 099 EBADCMD No such command\r\n");
    }

    if (db)
	dbmclose(db);

    return(0);
}

/* dcpd.c ends here */

