/**************************************************************************

NAME
   transact.c -- transaction machinery for DCP support library

DESCRIPTION
   Building blocks for DCP client protocol.  These expect to be able to call
an error() function resembling printf(3) for transation logging.

AUTHOR
   Eric S. Raymond <esr@thyrsus.com>

**************************************************************************/
#include "config.h"
#include <time.h>
#include "socket.h"
#include "dcp.h"
#include "dcplib.h"
#if defined(HAVE_UNISTD_H)
#include <unistd.h>
#endif
#if defined(HAVE_STDARG_H)
#include  <stdarg.h>
#else
#include  <varargs.h>
#endif
#include "string.h"

char *shroud;	/* password string to be shrouded, if any */
int verbose;	/* report transactions? */

#if defined(HAVE_STDARG_H)
void gen_send(int sock, const char *fmt, ... )
/* assemble command in printf(3) style and send to the server */
#else
void gen_send(sock, fmt, va_alist)
/* assemble command in printf(3) style and send to the server */
int sock;		/* socket to which server is connected */
const char *fmt;	/* printf-style format */
va_dcl
#endif
{
    char buf [DCPBUFSIZE+1];
    va_list ap;

#if defined(HAVE_STDARG_H)
    va_start(ap, fmt) ;
#else
    va_start(ap);
#endif
#ifdef HAVE_VSNPRINTF
    vsnprintf(buf + strlen(buf), sizeof(buf), fmt, ap);
#else
    vsprintf(buf + strlen(buf), fmt, ap);
#endif
    va_end(ap);

    strcat(buf, "\r\n");
    SockWrite(sock, buf, strlen(buf));

    if (verbose)
    {
	char *cp;

	if (shroud && shroud[0] && (cp = strstr(buf, shroud)))
	{
	    char	*sp;

	    sp = cp + strlen(shroud);
	    *cp++ = '*';
	    while (*sp)
		*cp++ = *sp++;
	    *cp = '\0';
	}
	buf[strlen(buf)-2] = '\0';
	error(0, 0, "> %s", buf);
    }
}

int gen_recv(sock, buf, size)
/* get one line of input from the server */
int sock;	/* socket to which server is connected */
char *buf;	/* buffer to receive input */
int size;	/* length of buffer */
{
    if (SockRead(sock, buf, size) == -1)
	return(DCP_ERR);
    else
    {
	if (buf[strlen(buf)-1] == '\n')
	    buf[strlen(buf)-1] = '\0';
	if (buf[strlen(buf)-1] == '\r')
	    buf[strlen(buf)-1] = '\r';
	if (verbose)
	{
	    char *cp;

	    if (shroud && shroud[0] && (cp = strstr(buf, shroud)))
	    {
		char	*sp;

		sp = cp + strlen(shroud);
		*cp++ = '*';
		while (*sp)
		    *cp++ = *sp++;
		*cp = '\0';
	    }
	    buf[strlen(buf)-1] = '\0';
	    error(0, 0, "< %s", buf);
	}

	if (strncmp(buf, "ERR ", 4) == 0)
	{
	    dcp_errno = atoi(buf + 4);
	    return(DCP_ERR);
	}
	else
	    return(DCP_OK);
    }
}

/* transact.c ends here */

